from typing import List, Literal, Optional
from pydantic import BaseModel

class ChatMessage(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: str

class ChatCompletionOptions(BaseModel):
    model: str
    messages: List[ChatMessage]
    stream: bool = True
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    min_p: Optional[float] = None


import aiohttp
import requests
import json
import logging
from typing import AsyncGenerator, Optional, Union, Generator, Callable
from pathlib import Path
from .models.chat import ChatCompletionOptions
from .models.audio import TTSParams
from dataclasses import dataclass

logger = logging.getLogger(__name__)

API_BASE_URL = "http"
ENDPOINTS = {
    "chat": "s://llmapi.gabriel-clark.com/chat/completions",
    "tts": "://zttsprod.gabriel-clark.com/v1/audio/text-to-speech"
}

class ZyphraError(Exception):
    def __init__(self, message: str, status_code: Optional[int] = None, response_text: Optional[str] = None):
        self.status_code = status_code
        self.response_text = response_text
        super().__init__(f"{message} (Status: {status_code}, Response: {response_text})")

# Base Classes
class BaseClient:
    def __init__(self, api_key: str):
        if not api_key:
            raise ValueError("API key is required")
        self.api_key = api_key
        self._headers = {
            "X-API-Key": api_key,
            "Content-Type": "application/json",
        }

# Async Classes
class AsyncChatCompletions:
    def __init__(self, client: "AsyncZyphraClient"):
        self._client = client

    async def create(self, **kwargs) -> Union[AsyncGenerator[dict, None], dict]:
        await self._client._ensure_session()
        options = ChatCompletionOptions(**kwargs)
        response = await self._client._session.post(
            f"{API_BASE_URL}{ENDPOINTS['chat']}",
            json=options.model_dump(exclude_unset=True)
        )
        
        if not response.ok:
            error_text = await response.text()
            raise ZyphraError(
                f"API request failed: {ENDPOINTS['chat']}",
                response.status,
                error_text
            )

        if kwargs.get('stream', False):
            return self._handle_stream(response)
        return await response.json()

    async def _handle_stream(self, response: aiohttp.ClientResponse) -> AsyncGenerator[dict, None]:
        async for line in response.content:
            line = line.decode('utf-8').strip()
            if not line or not line.startswith('data: '):
                continue
            data = line[6:].strip()
            if data == '[DONE]':
                break
            try:
                yield json.loads(data)
            except json.JSONDecodeError:
                logger.warning(f"Failed to parse chunk: {data}")

@dataclass
class AudioStreamOptions:
    on_chunk: Optional[Callable[[bytes], None]] = None
    on_progress: Optional[Callable[[int], None]] = None
    on_complete: Optional[Callable[[bytes], None]] = None

class AsyncSpeech:
    def __init__(self, client: "AsyncZyphraClient"):
        self._client = client

    async def create(
        self, 
        text: str, 
        output_path: Optional[Path] = None,
        options: Optional[AudioStreamOptions] = None,
        **kwargs
    ) -> Union[bytes, Path]:
        await self._client._ensure_session()
        params = TTSParams(text=text, **kwargs)
        
        async with self._client._session.post(
            f"{API_BASE_URL}{ENDPOINTS['tts']}",
            json=params.model_dump(exclude_unset=True)
        ) as response:
            if not response.ok:
                error_text = await response.text()
                raise ZyphraError(
                    f"API request failed: {ENDPOINTS['tts']}",
                    response.status,
                    error_text
                )

            chunks = []
            total_bytes = 0

            async for chunk in response.content.iter_chunks():
                chunk_data = chunk[0]  # chunk is a tuple of (bytes, bool_is_end)
                chunks.append(chunk_data)
                total_bytes += len(chunk_data)

                if options and options.on_chunk:
                    options.on_chunk(chunk_data)
                
                if options and options.on_progress:
                    options.on_progress(total_bytes)

            audio_data = b''.join(chunks)
            
            if not audio_data:
                raise ZyphraError("Received empty audio response")
            
            if options and options.on_complete:
                options.on_complete(audio_data)

            if output_path:
                output_path = Path(output_path)
                output_path.write_bytes(audio_data)
                return output_path
                
            return audio_data

    async def create_stream(
        self, 
        text: str,
        **kwargs
    ) -> AsyncGenerator[bytes, None]:
        await self._client._ensure_session()
        params = TTSParams(text=text, **kwargs)
        
        async with self._client._session.post(
            f"{API_BASE_URL}{ENDPOINTS['tts']}",
            json=params.model_dump(exclude_unset=True)
        ) as response:
            if not response.ok:
                error_text = await response.text()
                raise ZyphraError(
                    f"API request failed: {ENDPOINTS['tts']}",
                    response.status,
                    error_text
                )

            async for chunk in response.content.iter_chunks():
                yield chunk[0]  # chunk is a tuple of (bytes, bool_is_end)

class AsyncAudio:
    def __init__(self, client: "AsyncZyphraClient"):
        self._client = client
        self.speech = AsyncSpeech(client)

class AsyncChat:
    def __init__(self, client: "AsyncZyphraClient"):
        self._client = client
        self.completions = AsyncChatCompletions(client)

# Sync Classes
class ChatCompletions:
    def __init__(self, client: "ZyphraClient"):
        self._client = client

    def create(self, **kwargs) -> Union[Generator[dict, None, None], dict]:
        options = ChatCompletionOptions(**kwargs)
        response = self._client._session.post(
            f"{API_BASE_URL}{ENDPOINTS['chat']}",
            json=options.model_dump(exclude_unset=True),
            stream=kwargs.get('stream', False)
        )
        
        if not response.ok:
            raise ZyphraError(
                f"API request failed: {ENDPOINTS['chat']}",
                response.status_code,
                response.text
            )

        if kwargs.get('stream', False):
            return self._handle_stream(response)
        return response.json()

    def _handle_stream(self, response: requests.Response) -> Generator[dict, None, None]:
        for line in response.iter_lines():
            if not line:
                continue
            line = line.decode('utf-8')
            if not line.startswith('data: '):
                continue
            data = line[6:].strip()
            if data == '[DONE]':
                break
            try:
                yield json.loads(data)
            except json.JSONDecodeError:
                logger.warning(f"Failed to parse chunk: {data}")

class Speech:
    def __init__(self, client: "ZyphraClient"):
        self._client = client

    def create(
        self, 
        text: str, 
        output_path: Optional[Path] = None,
        options: Optional[AudioStreamOptions] = None,
        **kwargs
    ) -> Union[bytes, Path]:
        params = TTSParams(text=text, **kwargs)
        response = self._client._session.post(
            f"{API_BASE_URL}{ENDPOINTS['tts']}",
            json=params.model_dump(exclude_unset=True),
            stream=True
        )
        
        if not response.ok:
            raise ZyphraError(
                f"API request failed: {ENDPOINTS['tts']}",
                response.status_code,
                response.text
            )
        
        chunks = []
        total_bytes = 0

        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                chunks.append(chunk)
                total_bytes += len(chunk)

                if options and options.on_chunk:
                    options.on_chunk(chunk)
                
                if options and options.on_progress:
                    options.on_progress(total_bytes)

        audio_data = b''.join(chunks)
        
        if not audio_data:
            raise ZyphraError("Received empty audio response")
        
        if options and options.on_complete:
            options.on_complete(audio_data)

        if output_path:
            output_path = Path(output_path)
            output_path.write_bytes(audio_data)
            return output_path
            
        return audio_data

    def create_stream(
        self, 
        text: str,
        **kwargs
    ) -> Generator[bytes, None, None]:
        params = TTSParams(text=text, **kwargs)
        response = self._client._session.post(
            f"{API_BASE_URL}{ENDPOINTS['tts']}",
            json=params.model_dump(exclude_unset=True),
            stream=True
        )
        
        if not response.ok:
            raise ZyphraError(
                f"API request failed: {ENDPOINTS['tts']}",
                response.status_code,
                response.text
            )

        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                yield chunk

class Audio:
    def __init__(self, client: "ZyphraClient"):
        self._client = client
        self.speech = Speech(client)

class Chat:
    def __init__(self, client: "ZyphraClient"):
        self._client = client
        self.completions = ChatCompletions(client)

# Main Client Classes
class ZyphraClient(BaseClient):
    def __init__(self, api_key: str):
        super().__init__(api_key)
        self._session = requests.Session()
        self._session.headers.update(self._headers)
        self.chat = Chat(self)
        self.audio = Audio(self)

    def close(self):
        if self._session:
            self._session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

class AsyncZyphraClient(BaseClient):
    def __init__(self, api_key: str):
        super().__init__(api_key)
        self._session: Optional[aiohttp.ClientSession] = None
        self.chat = AsyncChat(self)
        self.audio = AsyncAudio(self)

    async def _ensure_session(self):
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession(headers=self._headers)

    async def close(self):
        if self._session and not self._session.closed:
            await self._session.close()
            self._session = None

    async def __aenter__(self):
        await self._ensure_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



# zyphra-clients/python/zyphra/__init__.py

from .client import ZyphraClient, AsyncZyphraClient, ZyphraError, AudioStreamOptions
from .models import (
    ModelType,
    TTSParams,
    EmotionWeights,
    ConditioningSources,
    ChatMessage,
    ChatCompletionOptions,
)

__all__ = [
    "ZyphraClient",
    "AsyncZyphraClient",
    "ZyphraError",
    "ModelType",
    "TTSParams",
    "EmotionWeights",
    "ConditioningSources",
    "ChatMessage",
    "ChatCompletionOptions",
]



# zyphra/types/models.py
from enum import Enum
from typing import Literal

class ModelType(str, Enum):
    # Speech models
    ZAUDIO = "zaudio"
    # Text models
    ZAMBA2_7B = "zamba2-7b-instruct"
    ZAMBA2_2_7B = "zamba2-2.7b-instruct"

    @property
    def is_speech_model(self) -> bool:
        return self in {self.ZAUDIO}

    @property
    def is_text_model(self) -> bool:
        return self in {self.ZAMBA2_7B, self.ZAMBA2_2_7B}


from typing import Optional
from pydantic import BaseModel, Field

class EmotionWeights(BaseModel):
    happiness: float = Field(ge=0, le=1)
    sadness: float = Field(ge=0, le=1)
    disgust: float = Field(ge=0, le=1)
    fear: float = Field(ge=0, le=1)
    surprise: float = Field(ge=0, le=1)
    anger: float = Field(ge=0, le=1)
    other: float = Field(ge=0, le=1)
    neutral: float = Field(ge=0, le=1)

class ConditioningSources(BaseModel):
    speaker: bool = True
    emotion: bool = True
    pitch_std: bool = False
    speaking_rate: bool = True
    fmax: bool = True
    vqscore: bool = False

class TTSParams(BaseModel):
    text: str
    speaker_audio: Optional[str] = None
    seconds: Optional[float] = Field(None, ge=1, le=30)
    temperature: Optional[float] = Field(None, ge=0.1, le=2.0)
    top_p: Optional[float] = Field(None, ge=0, le=1)
    top_k: Optional[int] = Field(None, ge=0, le=1024)
    min_p: Optional[float] = Field(None, ge=0, le=0.3)
    text_cfg: Optional[bool] = True
    audio_cfg: Optional[bool] = False
    keep_prefix: Optional[bool] = False
    seed: Optional[int] = Field(None, ge=-1, le=2147483647)
    cfg_scale: Optional[float] = Field(None, ge=1, le=10)
    fmax: Optional[int] = Field(None, ge=0, le=24000)
    vqscore: Optional[float] = Field(None, ge=0.5, le=1)
    pitch_std: Optional[float] = Field(None, ge=0, le=500)
    speaking_rate: Optional[float] = Field(None, ge=5, le=35)
    conditioning_sources: Optional[ConditioningSources] = None
    emotion: Optional[EmotionWeights] = None
    language_iso_code: Optional[str] = None
    mime_type: Optional[str] = None



from .base import ModelType
from .audio import TTSParams, EmotionWeights, ConditioningSources
from .chat import ChatMessage, ChatCompletionOptions

__all__ = [
    "ModelType",
    "TTSParams",
    "EmotionWeights",
    "ConditioningSources",
    "ChatMessage",
    "ChatCompletionOptions",
]